
// Schedules for reduction of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "TreeSchedule.h"
#include "Reduce.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce: reduction of array to scalar.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Reduce :: Reduce(const DAD* _src) : src(_src), rank(_src->rnk()) {
  Group p = src->grp() ;
  restrictToAPG(p) ;

  spanTree(&bcstTree, apg) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  if(p.member()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      spanTree(this, reduced) ;
  }
}

Reduce :: ~Reduce() {
  delete [] steps ;
  delete [] exts ;
}


void Reduce :: loop(const int offset, const int d) {
  if(d == rank)
    opReduce(offset) ;
  else {
    LocBlocksIndex i(src->rng(d)) ;

    Stride str = src->str(d) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int step = str.step(i.sub_stp) ;

      exts  [d] = i.count ;
      steps [d] = step ;

      loop(offset + str.offset(i), d + 1) ;
    }
  }
}


void Reduce :: execute(const int resLen, void* res) {
  Group p = src->grp() ;

  if(p.member()) {
    opZero() ;

    loop(0, 0) ;

    RedxSchedule :: combine(resLen, res) ;
  }

  bcstTree.broadcast(resLen, res, p.lead_abs()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce2: reduction of two arrays to scalar.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Reduce2 :: Reduce2(const DAD* _src1, const DAD* _src2) :
              src1(_src1), src2(_src2), rank(_src1->rnk()) {
  Group p = src1->grp() ;
  restrictToAPG(p) ;

  spanTree(&bcstTree, apg) ;

  exts   = new int [rank] ;
  steps1 = new int [rank] ;
  steps2 = new int [rank] ;

  if(p.member()) {
    DimensionSet rep = p.dims() - src1->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      spanTree(this, reduced) ;
  }
}

Reduce2 :: ~Reduce2() {
  delete [] steps2 ;
  delete [] steps1 ;
  delete [] exts ;
}


void Reduce2 :: loop(const int offset1, const int offset2, const int d) {
  if(d == rank)
    opReduce(offset1, offset2) ;
  else {
    LocBlocksIndex i(src1->rng(d)) ;

    Stride str1 = src1->str(d) ;
    Stride str2 = src2->str(d) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int step1 = str1.step(i.sub_stp) ;
      const int step2 = str2.step(i.sub_stp) ;

      exts   [d] = i.count ;
      steps1 [d] = step1 ;
      steps2 [d] = step2 ;

      loop(offset1 + str1.offset(i), offset2 + str2.offset(i), d + 1) ;
    }
  }
}


void Reduce2 :: execute(const int resLen, void* res) {
  Group p = src1->grp() ;

  if(p.member()) {
    opZero() ;

    loop(0, 0, 0) ;

    RedxSchedule :: combine(resLen, res) ;
  }

  bcstTree.broadcast(resLen, res, p.lead_abs()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim: Dimensional reduction of array.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReduceDim :: ReduceDim(const DAD* _res, const int len,
                       const DAD* _src, const int _dim) :
                 BlockRedxSchedule(len),
		 res(_res), src(_src), dim(_dim),
                 resRank(_res->rnk()), srcRank(_src->rnk()) {
  exts = new int [srcRank] ;
  srcSteps = new int [srcRank] ;
  resSteps = new int [resRank] ;

  Group p = res->grp() ;
  Range x = src->rng(dim) ;
  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    spanTree(this, reduced) ;
  }
}

ReduceDim :: ~ReduceDim() {
  delete [] resSteps ;
  delete [] srcSteps ;
  delete [] exts ;
}

void ReduceDim :: loop(const int resOffset, const int resCtg,
                       const int srcOffset,
		       const int vol, const int d, const int r) {
  if(d == srcRank) {
    LocBlocksIndex i(src->rng(dim)) ;

    Stride srcStr = src->str(dim) ;

    opZero(resOffset) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int srcStep = srcStr.step(i.sub_stp) ;

      exts     [resRank] = i.count ;
      srcSteps [resRank] = srcStep ;

      opReduce(resOffset, srcOffset + srcStr.offset(i)) ;
    }

    BlockRedxSchedule :: dataReq(resOffset, resSteps, resCtg, vol,
                                 exts, resRank) ;
  }
  else
    if(d == dim)
      loop(resOffset, resCtg, srcOffset, vol, d + 1, r) ;
    else {
      LocBlocksIndex i(src->rng(d)) ;

      Stride srcStr = src->str(d) ;
      Stride resStr = res->str(r) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

        const int srcStep = srcStr.step(i.sub_stp) ;
        const int resStep = resStr.step(i.sub_stp) ;

	exts     [r] = i.count ;
	srcSteps [r] = srcStep ;
	resSteps [r] = resStep ;

	loop(resOffset + resStr.offset(i), resCtg && resStep == vol,
             srcOffset + srcStr.offset(i),
	     vol * i.count, d + 1, r + 1) ;
      }
    }
}

void ReduceDim :: execute(void* resBase) {
  if(src->grp().member()) {
    loop(0, 1, 0, 1, 0, 0) ;

    BlockRedxSchedule :: combine(resBase) ;

    BlockTreeSchedule :: clear() ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim2: Dimensional reduction of two arrays.             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReduceDim2 :: ReduceDim2(const DAD* _res, const int len,
                         const DAD* _src1, const DAD* _src2,
                         const int _dim) :
                 BlockRedxSchedule(len),
		 res(_res), src1(_src1), src2(_src2), dim(_dim),
                 resRank(_res->rnk()), srcRank(_src1->rnk()) {
  exts = new int [srcRank] ;
  src1Steps = new int [srcRank] ;
  src2Steps = new int [srcRank] ;
  resSteps  = new int [resRank] ;

  Group p = res->grp() ;
  Range x = src1->rng(dim) ;
  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    spanTree(this, reduced) ;
  }
}

ReduceDim2 :: ~ReduceDim2() {
  delete [] resSteps ;
  delete [] src2Steps ;
  delete [] src1Steps ;
  delete [] exts ;
}

void ReduceDim2 :: loop(const int resOffset, const int resCtg,
                        const int src1Offset, const int src2Offset,
		        const int vol, const int d, const int r) {
  if(d == srcRank) {
    LocBlocksIndex i(src1->rng(dim)) ;

    Stride src1Str = src1->str(dim) ;
    Stride src2Str = src2->str(dim) ;

    opZero(resOffset) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int src1Step = src1Str.step(i.sub_stp) ;
      const int src2Step = src2Str.step(i.sub_stp) ;

      exts [resRank]      = i.count ;
      src1Steps [resRank] = src1Step ;
      src2Steps [resRank] = src2Step ;

      opReduce(resOffset, src1Offset + src1Str.offset(i),
                          src2Offset + src2Str.offset(i)) ;
    }

    BlockRedxSchedule :: dataReq(resOffset, resSteps, resCtg, vol,
                                 exts, resRank) ;
  }
  else
    if(d == dim)
      loop(resOffset, resCtg, src1Offset, src2Offset, vol, d + 1, r) ;
    else {
      LocBlocksIndex i(src1->rng(d)) ;

      Stride src1Str = src1->str(d) ;
      Stride src2Str = src2->str(d) ;
      Stride resStr  = res->str(r) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

        const int src1Step = src1Str.step(i.sub_stp) ;
        const int src2Step = src2Str.step(i.sub_stp) ;
        const int resStep  = resStr.step(i.sub_stp) ;

	exts [r]      = i.count ;
	src1Steps [r] = src1Step ;
	src2Steps [r] = src2Step ;
	resSteps  [r] = resStep ;

	loop(resOffset + resStr.offset(i), resCtg && resStep == vol,
             src1Offset + src1Str.offset(i), src2Offset + src2Str.offset(i),
	     vol * i.count, d + 1, r + 1) ;
      }
    }
}

void ReduceDim2 :: execute(void* resBase) {
  if(src1->grp().member()) {
    loop(0, 1, 0, 0, 1, 0, 0) ;

    BlockRedxSchedule :: combine(resBase) ;

    BlockTreeSchedule :: clear() ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate: reduction of array to scalar, tagging elements by    //
//                position in source array.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Locate :: Locate(const DAD* _src) : src(_src), rank(_src->rnk()) {
  Group p = src->grp() ;
  restrictToAPG(p) ;

  spanTree(&bcstTree, apg) ;

  exts  = new int [rank] ;
  srcSteps   = new int [rank] ;
  blockBases = new int [rank] ;
  blockSteps = new int [rank] ;

  if(p.member()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      spanTree(this, reduced) ;
  }
}

Locate :: ~Locate() {
  delete [] blockSteps ;
  delete [] blockBases ;
  delete [] srcSteps   ;
  delete [] exts ;
}


void Locate :: loop(const int offset, const int d) {
  if(d == rank)
    opReduce(offset) ;
  else {
    LocBlocksIndex i(src->rng(d)) ;
    Stride str = src->str(d) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int step = str.step(i.sub_stp) ;

      exts  [d] = i.count ;
      srcSteps   [d] = step ;
      blockBases [d] = i.glb_bas ;
      blockSteps [d] = i.glb_stp ;

      loop(offset + str.offset(i), d + 1) ;
    }
  }
}


void Locate :: execute(const int valLen, void* val, int* tags) {
  Group p = src->grp() ;

  if(p.member()) {
    opZero() ;
    loop(0, 0) ;
  }

  const int tagsSize = rank * sizeof(int) ;

  Redx2Schedule :: combine(valLen, val, tagsSize, tags) ;
  bcstTree.broadcast(valLen, val, tagsSize, tags, p.lead_abs()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate2: reduction of two arrays to scalar, tagging elements //
//                 by position in source arrays.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Locate2 :: Locate2(const DAD* _src1, const DAD* _src2) :
               src1(_src1), src2(_src2), rank(_src1->rnk()) {
  Group p = src1->grp() ;
  restrictToAPG(p) ;

  spanTree(&bcstTree, apg) ;

  exts   = new int [rank] ;
  src1Steps  = new int [rank] ;
  src2Steps  = new int [rank] ;
  blockBases = new int [rank] ;
  blockSteps = new int [rank] ;

  if(p.member()) {
    DimensionSet rep = p.dims() - src1->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      spanTree(this, reduced) ;
  }
}

Locate2 :: ~Locate2() {
  delete [] blockSteps ;
  delete [] blockBases ;
  delete [] src2Steps ;
  delete [] src1Steps ;
  delete [] exts ;
}


void Locate2 :: loop(const int offset1, const int offset2, const int d) {
  if(d == rank)
    opReduce(offset1, offset2) ;
  else {
    LocBlocksIndex i(src1->rng(d)) ;
    Stride str1 = src1->str(d) ;
    Stride str2 = src2->str(d) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int step1 = str1.step(i.sub_stp) ;
      const int step2 = str2.step(i.sub_stp) ;

      exts   [d] = i.count ;
      src1Steps [d]  = step1 ;
      src2Steps [d]  = step2 ;
      blockBases [d] = i.glb_bas ;
      blockSteps [d] = i.glb_stp ;

      loop(offset1 + str1.offset(i), offset2 + str2.offset(i), d + 1) ;
    }
  }
}


void Locate2 :: execute(const int valLen, void* val, int* tags) {
  Group p = src1->grp() ;

  if(p.member()) {
    opZero() ;
    loop(0, 0, 0) ;
  }

  const int tagsSize = rank * sizeof(int) ;

  Redx2Schedule :: combine(valLen, val, tagsSize, tags) ;
  bcstTree.broadcast(valLen, val, tagsSize, tags, p.lead_abs()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim: Dimensional reduction of array, tagging elements  //
//                   by position in reduced dimension of source array. //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

LocateDim :: LocateDim(const DAD* _val, const int len, const DAD* _tag,
                       const DAD* _src, const int _dim) :
                 BlockRedx2Schedule(len, sizeof(int)),
		 val(_val), tag(_tag), src(_src), dim(_dim),
                 resRank(_val->rnk()), srcRank(_src->rnk()) {
  exts     = new int [srcRank] ;
  srcSteps = new int [srcRank] ;
  valSteps = new int [resRank] ;
  tagSteps = new int [resRank] ;

  Group p = val->grp() ;
  Range x = src->rng(dim) ;
  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    spanTree(this, reduced) ;
  }
}

LocateDim :: ~LocateDim() {
  delete [] tagSteps ;
  delete [] valSteps ;
  delete [] srcSteps ;
  delete [] exts ;
}

void LocateDim :: loop(const int valOffset, const int valCtg,
		       const int tagOffset, const int tagCtg,
		       const int srcOffset,
		       const int vol, const int d, const int r) {

  if(d == srcRank) {
    LocBlocksIndex i(src->rng(dim)) ;

    Stride srcStr = src->str(dim) ;

    opZero(valOffset, tagOffset) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int srcStep = srcStr.step(i.sub_stp) ;

      exts     [resRank] = i.count ;
      srcSteps [resRank] = srcStep ;

      opReduce(valOffset, tagOffset,
	       srcOffset + srcStr.offset(i), i.glb_bas, i.glb_stp) ;
    }

    BlockRedx2Schedule :: dataReq(valOffset, valSteps, valCtg, vol, 
				  tagOffset, tagSteps, tagCtg, vol,
				  exts, resRank) ;
  }
  else
    if(d == dim)
      loop(valOffset, valCtg, tagOffset, tagCtg, srcOffset, vol, d + 1, r) ;
    else {
      LocBlocksIndex i(src->rng(d)) ;

      Stride srcStr = src->str(d) ;
      Stride valStr = val->str(r) ;
      Stride tagStr = tag->str(r) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	const int srcStep = srcStr.step(i.sub_stp) ;
	const int valStep = valStr.step(i.sub_stp) ;
	const int tagStep = tagStr.step(i.sub_stp) ;

	exts     [r] = i.count ;
	srcSteps [r] = srcStep ;
	valSteps [r] = valStep ;
	tagSteps [r] = tagStep ;

	loop(valOffset + valStr.offset(i), valCtg && valStep == vol,
	     tagOffset + tagStr.offset(i), tagCtg && tagStep == vol,
	     srcOffset + srcStr.offset(i),
	     vol * i.count, d + 1, r + 1) ;
      }
    }
}

void LocateDim :: execute(void* valBase, int* tagBase) {
  if(src->grp().member())
    loop(0, 1, 0, 1, 0, 1, 0, 0) ;

  BlockRedx2Schedule :: combine(valBase, tagBase) ;

  BlockTreeSchedule :: clear() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim2: Dimensional reduction of two arrays, tagging     //
//                    elements by position in reduced dimension of     //
//                    source arrays.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

LocateDim2 :: LocateDim2(const DAD* _val, const int len, const DAD* _tag,
                         const DAD* _src1, const DAD* _src2,
                         const int _dim) :
                 BlockRedx2Schedule(len, sizeof(int)),
		 val(_val), tag(_tag), src1(_src1), src2(_src2), dim(_dim),
                 resRank(_val->rnk()), srcRank(_src1->rnk()) {
  exts      = new int [srcRank] ;
  src1Steps = new int [srcRank] ;
  src2Steps = new int [srcRank] ;
  valSteps  = new int [resRank] ;
  tagSteps  = new int [resRank] ;

  Group p = val->grp() ;
  Range x = src1->rng(dim) ;
  if(p.member()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'

    Group reduced = p ;
    reduced.restrict(orth) ;  // Includes this process, and spanned by x.dim()

    spanTree(this, reduced) ;
  }
}

LocateDim2 :: ~LocateDim2() {
  delete [] tagSteps  ;
  delete [] valSteps  ;
  delete [] src1Steps ;
  delete [] src2Steps ;
  delete [] exts ;
}

void LocateDim2 :: loop(const int valOffset, const int valCtg,
		        const int tagOffset, const int tagCtg,
		        const int src1Offset, const int src2Offset,
		        const int vol, const int d, const int r) {

  if(d == srcRank) {
    LocBlocksIndex i(src1->rng(dim)) ;

    Stride src1Str = src1->str(dim) ;
    Stride src2Str = src2->str(dim) ;

    opZero(valOffset, tagOffset) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int src1Step = src1Str.step(i.sub_stp) ;
      const int src2Step = src2Str.step(i.sub_stp) ;

      exts [resRank]      = i.count ;
      src1Steps [resRank] = src1Step ;
      src2Steps [resRank] = src2Step ;

      opReduce(valOffset, tagOffset,
	       src1Offset + src1Str.offset(i), src2Offset + src2Str.offset(i),
               i.glb_bas, i.glb_stp) ;
    }

    BlockRedx2Schedule :: dataReq(valOffset, valSteps, valCtg, vol, 
				  tagOffset, tagSteps, tagCtg, vol,
				  exts, resRank) ;
  }
  else
    if(d == dim)
      loop(valOffset, valCtg, tagOffset, tagCtg, src1Offset, src2Offset,
           vol, d + 1, r) ;
    else {
      LocBlocksIndex i(src1->rng(d)) ;

      Stride src1Str = src1->str(d) ;
      Stride src2Str = src2->str(d) ;
      Stride valStr  = val->str(r) ;
      Stride tagStr  = tag->str(r) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	const int src1Step = src1Str.step(i.sub_stp) ;
	const int src2Step = src2Str.step(i.sub_stp) ;
	const int valStep  = valStr.step(i.sub_stp) ;
	const int tagStep  = tagStr.step(i.sub_stp) ;

	exts [r]      = i.count ;
	src1Steps [r] = src1Step ;
	src2Steps [r] = src2Step ;
	valSteps  [r] = valStep ;
	tagSteps  [r] = tagStep ;

	loop(valOffset + valStr.offset(i), valCtg && valStep == vol,
	     tagOffset + tagStr.offset(i), tagCtg && tagStep == vol,
	     src1Offset + src1Str.offset(i), src2Offset + src2Str.offset(i),
	     vol * i.count, d + 1, r + 1) ;
      }
    }
}

void LocateDim2 :: execute(void* valBase, int* tagBase) {
  if(src1->grp().member())
    loop(0, 1, 0, 1, 0, 0, 1, 0, 0) ;

  BlockRedx2Schedule :: combine(valBase, tagBase) ;

  BlockTreeSchedule :: clear() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Replicate: broadcast elements of an array from roots of the  //
//                   groups over which they are replicated.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Replicate :: Replicate(const DAD* _res, const int len) :
                 BlockTreeSchedule(len), res(_res), rank(_res->rnk()) {
  exts  = new int [rank] ;
  steps = new int [rank] ;

  Group p = res->grp() ;
  if(p.member()) {
    loop(0, 1, 1, 0) ;

    // Computing spanning tree of replication group.

    Group reduced = p ;
    reduced.restrict(res->sig()) ;

    spanTree(this, reduced) ;
  }
}

Replicate :: ~Replicate() {
  delete [] steps ;
  delete [] exts ;
}

void Replicate :: loop(const int offset, const int ctg,
		       const int vol, const int d) {
  if(d == rank)
    BlockTreeSchedule :: dataReq(offset, steps, ctg, vol, exts, rank) ;
  else {
    LocBlocksIndex i(res->rng(d)) ;

    Stride str = res->str(d) ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      const int step = str.step(i.sub_stp) ;

      exts  [d] = i.count ;
      steps [d] = step ;

      loop(offset + str.offset(i), ctg && step == vol,
	   vol * i.count, d + 1) ;
    }
  }
}

void Replicate :: execute(void* resBase) {
  BlockTreeSchedule :: broadcast(resBase) ;
}


// ----------------------------------------------------------------------
// Things to do
//
//     Where there are several short vectors in a schedule (`exts',
//     `steps', ...) allocate and deallocate them as a single vector.
//
//     Or allocate them using `Vec'?
//
//     Schedules should contain vectors of `LocBlocksIndex'?
//
//     Many declaration of the type...
//
//       const int step = str.step(i.sub_stp) ;
//
//     ...are redundant, because results aren't reused.  Put the expressions
//     inline?
//
//     Think of a better name for `dataReq'.  (Probably just `req').
//
//     **** Shouldn't execute `loop' etc if spanning tree is trivial. ****

